from typing import override
import pandas as pd

from taker.TrailingSLTaker import TrailingSLTaker

class SMCSLAndTPTaker(TrailingSLTaker):
    def __init__(self,g_config, platform_config, feishu_webhook=None, monitor_interval=4,logger=None):
        super().__init__(g_config, platform_config, feishu_webhook, monitor_interval,logger)
        self.global_symbol_take_profit_flag = {} # 记录每个symbol是否设置全局止盈标志
        self.global_symbol_take_profit_price = {} # 记录每个symbol的止盈价格
        self.htf_liquidities = {}
        self.all_TP_SL_ratio = float(platform_config.get("all_TP_SL_ratio",1.5)) #The profit-loss ratio 盈亏比
        self.open_trail_profit = bool(platform_config.get("open_trail_profit",True)) # 开仓是否设置止盈
       
    @override  
    def check_reverse_position(self,symbol,position,pair_config):
        """
        检查是否有反向持仓
        """
        self.logger.debug(f"{symbol}: 检查LTF-Struceture是否市价清仓。")
        # 根据LTF的Struceture识别
        
 
     
    def build_struct(self, df, prd=20, check_bounds=True, global_extremum=False) :
        
        """_summary_
            构建SMC结构，参考 Tradingview Smart Money Concepts Probability (Expo)@Openfund
        """
        data = df.copy()
        data['Up'] = None
        data['Dn'] = None
        data['iUp'] = None
        data['iDn'] = None
        data['pos'] = 0
        data['pattern'] = None

        # 初始化 Up 和 Dn 的第一个值
        data.at[0, 'Up'] = data.at[0, 'high']
        data.at[0, 'Dn'] = data.at[0, 'low']
        

        for index in range(1, len(data)):
      
            data.at[index, 'Up'] = max(data.at[index - 1, 'Up'], data.at[index, 'high'])
            data.at[index, 'Dn'] = min(data.at[index - 1, 'Dn'], data.at[index, 'low'])
            data.at[index, 'pos'] = data.at[index - 1, 'pos']
            data.at[index, 'iUp'] = data.at[max(0,index - 1), 'iUp'] if data.at[max(0,index - 1), 'iUp'] is not None else index
            data.at[index, 'iDn'] = data.at[max(0,index - 1), 'iDn'] if data.at[max(0,index - 1), 'iDn'] is not None else index

            # 寻找枢轴高点和低点
            pvtHi = self.is_pivot_high(data, index, prd, check_bounds)
            pvtLo = self.is_pivot_low(data, index, prd, check_bounds)

            if pvtHi:
                data.at[index, 'Up'] = data.at[index, 'high']
                data.at[index, 'iUp'] = index
            if pvtLo:
                data.at[index, 'Dn'] = data.at[index, 'low']
                data.at[index, 'iDn'] = index
            # 寻找Bullish结构
            if data.at[index, 'Up'] > data.at[index - 1, 'Up']:
                data.at[index, 'iUp'] = index # TODO
                if data.at[index - 1, 'pos'] <= 0:
                    # data.at[index, 'pattern'] = 'CHoCH (Bullish)'
                    data.at[index, 'pattern'] = 'Bullish_CHoCH'
                    data.at[index, 'pos'] = 1
                elif data.at[index - 1, 'pos'] == 1 \
                        and data.at[index - 1, 'Up'] == data.at[max(0,index - prd), 'Up']:               
                    data.at[index, 'pattern'] = 'Bullish_SMS'
                    data.at[index, 'pos'] = 2
                    
                elif data.at[index - 1, 'pos'] > 1 \
                        and data.at[index - 1, 'Up'] == data.at[max(0,index - prd), 'Up']:                
                    data.at[index, 'pattern'] = 'Bullish_BMS'
                    data.at[index, 'pos'] = data.at[index - 1, 'pos'] + 1
                    
            elif global_extremum and data.at[index, 'Up'] < data.at[index - 1, 'Up']:
                data.at[index, 'iUp'] = data.at[index - 1, 'iUp']
        
            # # 寻找Bearish结构
            if data.at[index, 'Dn'] < data.at[index - 1, 'Dn']:
                data.at[index, 'iDn'] = index # TODO
                if data.at[index - 1, 'pos'] >= 0:
                
                    data.at[index, 'pattern'] = 'Bearish_CHoCH'
                    data.at[index, 'pos'] = -1
                elif data.at[index - 1, 'pos'] == -1  \
                        and data.at[index - 1, 'Dn'] == data.at[max(0,index - prd), 'Dn']:
                    data.at[index, 'pattern'] = 'Bearish_SMS'
                    data.at[index, 'pos'] = -2
                elif data.at[index - 1, 'pos'] < -1  \
                        and data.at[index - 1, 'Dn'] == data.at[max(0,index - prd), 'Dn']:
                    data.at[index, 'pattern'] = 'Bearish_BMS'
                    data.at[index, 'pos'] = data.at[index - 1, 'pos'] - 1
                    
            elif global_extremum and data.at[index, 'Dn'] > data.at[index - 1, 'Dn']:
                data.at[index, 'iDn'] = data.at[index - 1, 'iDn']
                
        return data
            
    def detect_struct(self, data, prd=20, check_valid_range=True, struct_key=None, check_bounds=True, global_extremum=False) -> dict:
        """_summary_    
            识别智能资金结构
        
        :param data: 包含 'high' 和 'low' 列的 DataFrame
        :param prd: 结构周期
        :param s1: 结构响应布尔值
        :param resp: 响应周期
        :return: 包含结构识别结果的 DataFrame
        """

        data = self.build_struct(data, prd, check_bounds, global_extremum)
               
        
        # 获取最后一个结构和位置
        last_struct = {
            "struct": None,
            "index": -1,
            "pivot_high": None,
            "pivot_high_index": -1,
            "pivot_low": None,
            "pivot_low_index": -1,
            "side": None
            
        }
        
        pivot_high_index = last_struct["pivot_high_index"] = int(data["iUp"].iloc[-1])
        pivot_low_index = last_struct["pivot_low_index"] = int(data["iDn"].iloc[-1])
        
        last_struct["pivot_high"] = float(data.loc[last_struct["pivot_high_index"], 'high'])
        last_struct["pivot_low"] = float(data.loc[last_struct["pivot_low_index"], 'low'])
        
        for i in range(len(data)-1, -1, -1):
            if check_valid_range:
                # 检查是否在pivot_high_index和pivot_low_index之间的有效范围内
                if data.at[i, 'iUp'] != -1 and data.at[i, 'iDn'] != -1:
                    # pivot_high_index = data.at[i, 'iUp'] 
                    # pivot_low_index = data.at[i, 'iDn']
                    if i < min(pivot_high_index, pivot_low_index) or i > max(pivot_high_index, pivot_low_index):
                        continue
            
            if data.at[i, 'pattern'] is not None:
                if struct_key is not None and struct_key not in data.at[i, 'pattern']:
                    continue
                last_struct["struct"] = data.at[i, 'pattern']
                last_struct["index"] = i
               
                break
        
        if last_struct['struct'] is not None :
            # 找到最后一个结构的枢轴高点和低点，如果当前是孤立点，则取前一个孤立点
            # 判断交易方向
            if 'Bearish' in last_struct["struct"]:
                last_struct["side"] = 'sell'
            else :
                last_struct["side"] = 'buy'
        else:
            last_struct['struct'] = 'None'
            last_struct["index"] = -1

            
        return last_struct
 
    def is_pivot_high(self, data, index, period, check_bounds=False):
        """
        判断当前索引处是否为枢轴高点
        :param data: 包含 'high' 列的 DataFrame
        :param index: 当前索引
        :param period: 前后比较的周期数
        :return: 是否为枢轴高点
        """
        if check_bounds and (index < period or index >= len(data) - period):
            return False
        current_high = data.at[index, 'high']
        prev_highs = data['high'].iloc[max(0,index - period):index]
        next_highs = data['high'].iloc[index :min(len(data),index + period + 1)]
        return all(current_high >= prev_highs) and all(current_high >= next_highs)

    def is_pivot_low(self, data, index, period, check_bounds=False):
        """
        判断当前索引处是否为枢轴低点
        :param data: 包含 'low' 列的 DataFrame
        :param index: 当前索引
        :param period: 前后比较的周期数
        :return: 是否为枢轴低点
        """
        if check_bounds and (index < period or index >= len(data) - period):
            return False
        current_low = data.at[index, 'low']
        prev_lows = data['low'].iloc[max(0,index - period):index]
        next_lows = data['low'].iloc[index :min(len(data),index + period + 1)]
        return all(current_low <= prev_lows) and all(current_low <= next_lows)

              
    def set_sl_by_profit(self, symbol, position, profit, pair_config, kLines=None):
        
        """
        根据利润设置止损价
        Args:
            symbol: 交易对
            position: 仓位信息
            profit: 利润
            pair_config: 交易对配置
            kLines: K线数据
        """

        total_profit = profit      

        current_tier = '无'
        # 各档止盈逻辑
        # 确定当前盈利档位
        if self.open_trail_profit and total_profit >= self.second_trail_profit_threshold:
            current_tier = "高档"
     
        elif self.open_trail_profit and total_profit>= self.first_trail_profit_threshold:
            current_tier = "中档"
         
        elif self.open_trail_profit and total_profit >= self.low_trail_profit_threshold:
            current_tier = "低档"
         
        # 根据不同档位设置止损价格,没有单独为交易对设置，用全局参数代替
        tier_config = {
            "低档": {
                "stop_loss_pct": float(pair_config.get('low_trail_stop_loss_pct',self.low_trail_stop_loss_pct)),
                         
            },
            "中档": {
                "stop_loss_pct": float(pair_config.get('trail_stop_loss_pct',self.trail_stop_loss_pct)),
            
            },
            "高档": {
                "stop_loss_pct": float(pair_config.get('higher_trail_stop_loss_pct',self.higher_trail_stop_loss_pct)),

            }
        }
        self.logger.info(
                f"{symbol} 档位[{current_tier}]: 盈利: {total_profit:.2f}%")

        if current_tier in tier_config:
            config = tier_config[current_tier]  

                 

            # 计算回撤止损价格
            sl_price = self.calculate_stop_loss_price(
                symbol=symbol, 
                position=position,
                stop_loss_pct=config['stop_loss_pct']
            )
            
            # 检查价格是否变化
            latest_sl_price = self.exchange.safe_float(self.global_symbol_stop_loss_price, symbol, 0.0)
            if sl_price == latest_sl_price:
                self.logger.debug(f"{symbol}: 回撤止损价格{latest_sl_price}未变化，不设置")
                return
            
            # 根据持仓方向判断是否需要更新止损价格
            should_update = False
            if position['side'] == 'long':
                # 多仓时,新止损价需高于当前止损价
                should_update = sl_price >= latest_sl_price
                if not should_update:
                    self.logger.debug(f"{symbol}: 多仓回撤止损价格{sl_price}低于当前止损价格{latest_sl_price}，不更新")
                else:
                    self.logger.debug(f"{symbol}: 多仓回撤止损价格{sl_price}高于当前止损价格{latest_sl_price}，更新")
            else:
                # 空仓时,新止损价需低于当前止损价
                should_update = sl_price <= latest_sl_price
                if not should_update:
                    self.logger.debug(f"{symbol}: 空仓回撤止损价格{sl_price}高于当前止损价格{latest_sl_price}，不更新")
                else:
                    self.logger.debug(f"{symbol}: 空仓回撤止损价格{sl_price}低于当前止损价格{latest_sl_price}，更新")
            
            if not should_update:
                return  
            
            self.cancel_all_algo_orders(symbol=symbol, attachType='SL')
                
            # 移动止损保护
            if_success =  self.set_stop_loss(symbol=symbol, position=position, stop_loss_price=sl_price)
            
            if if_success:
                # 更新回撤止损价格
         
                self.global_symbol_stop_loss_price[symbol] = sl_price
                self.global_symbol_stop_loss_flag[symbol] = True
                cur_highest_total_profit = self.highest_total_profit.get(symbol, 0.0) 
                
                # 发送通知
                msg = (f"{symbol}: 盈利达到【{current_tier}】阈值，最高总盈利: {cur_highest_total_profit:.2f}%,"
                      f"当前盈利回撤到: {total_profit:.2f}%，市场价格:{position['markPrice']},"
                      f"设置回撤止损位: {sl_price:.9f}")
                self.logger.info(msg)
                self.send_feishu_notification(msg)
                
        else:

            # 默认全局止损            
            if_success = self.set_global_stop_loss(symbol=symbol, position=position, pair_config=pair_config ,kLines = kLines)
            if if_success:
                # 更新回撤止损价格 
                self.global_symbol_take_profit_flag[symbol] = False
        
    def find_liquidity(self, symbol, data, liquidity_type="BSL") -> pd.DataFrame:
        """
        寻找流动性，根据side判断是做多还是做空，做多则寻找iUp，做空则寻找iDn
        Args:
            symbol (str): 交易对
            data (pd.DataFrame): 数据
            liquidity_type (str): 流动性类型，'BSL' 或 'BSL'
            
        """
        df = data.copy()
        
        is_buy = liquidity_type == 'BSL'
        col_prefix = 'iDn' if is_buy else 'iUp'
        
        return df[df.index == df[col_prefix]].sort_index(ascending=False)
    
    def detect_liquidity_for_TP(self, symbol, data, side , market_price) -> pd.DataFrame:
        """
        TP校对流动性，用市场价格校验流动性是否有效,做多则流动性在市场价格之上，做空流动性要在市场价格之下。
        Args:
            symbol (str): 交易对
            side (str): 交易方向，'long' 或 'short'
            df_liquidities (pd.DataFrame): 流动性数据
            market_price (float): 当前市场价格
            
        """
        is_buy = side == 'long'
        col_prefix = 'iUp' if is_buy else 'iDn'
        price_col = 'Up' if is_buy else 'Dn'
        
        # 设置TP,long寻找SSL,short寻找BSL
        liquidity_type = 'SSL' if is_buy else 'BSL'
        
        df_liquidities = self.find_liquidity(symbol, data, liquidity_type=liquidity_type)

        df_valid_liquidities = df_liquidities.copy()     
                 
        result_indices = []
        current_price = float('-inf') if is_buy else float('inf')
        current_i = float('inf') 
        
        # 遍历并筛选符合条件的记录
        for idx, row in df_valid_liquidities.iterrows():
            if is_buy:
                if  row[price_col] > current_price and row[price_col] > market_price and row[col_prefix] < current_i:
                    result_indices.append(idx)
                    current_price = row[price_col]
                    current_i = row[col_prefix]
            else:
                if  row[price_col] < current_price and row[price_col] < market_price and row[col_prefix] < current_i:
                    result_indices.append(idx)
                    current_price = row[price_col]
                    current_i = row[col_prefix]
                    
        return df_valid_liquidities.loc[result_indices].sort_index(ascending=False)
    
    def detect_liquidity_for_SL(self, symbol, data, side, market_price) -> pd.DataFrame:
        """
        SL校对流动性，用市场价格校验流动性是否有效,做多则流动性在市场价格之下，做空流动性要在市场价格之上。
        Args:
            symbol (str): 交易对
            side (str): 交易方向，'long' 或'short'
            df_liquidities (pd.DataFrame): 流动性数据
            market_price (float): 当前市场价格

        """
        is_buy = side == 'long'
        col_prefix = 'iDn' if is_buy else 'iUp'
        price_col = 'Dn' if is_buy else 'Up'
        # 设置SL,long寻找BSL,short寻找SSL
        liquidity_type = 'BSL' if is_buy else 'SSL'
        
        df_liquidities = self.find_liquidity(symbol, data, liquidity_type=liquidity_type)

        df_valid_liquidities = df_liquidities.copy()     
        df_valid_liquidities['was_swept'] = None  
        df_valid_liquidities['took_out'] = None   
        
        result_indices = []
        current_price = float('inf') if is_buy else float('-inf')
        current_i = float('inf') 
        took_out_status = False
        
        # 遍历并筛选符合条件的记录，找到扫过流动性的
        for idx, row in df_valid_liquidities.iterrows():
            if not is_buy:
                if row[price_col] > market_price:
                    if  row[price_col] > current_price and row[col_prefix] < current_i:
                        result_indices.append(idx)
                        
                        if current_i != float('inf')  and  not df_valid_liquidities.at[current_i, 'took_out'] :
                            df_valid_liquidities.at[idx, 'took_out'] = True
                            
                        current_price = row[price_col]
                        current_i = row[col_prefix]
                        
                    else:
                        df_valid_liquidities.at[idx, 'was_swept'] = True
               
            else:
                if row[price_col] < market_price:
                    if  row[price_col] < current_price and row[col_prefix] < current_i:
                        result_indices.append(idx)
                        
                        if current_i != float('inf')  and  not df_valid_liquidities.at[current_i, 'took_out'] :
                            df_valid_liquidities.at[idx, 'took_out'] = True
                        current_price = row[price_col]
                        current_i = row[col_prefix]
                    else:
                        df_valid_liquidities.at[idx, 'was_swept'] = True
                        
        # 过滤出took_out为True的记录
        took_out_df = df_valid_liquidities.loc[result_indices]
        took_out_df = took_out_df[took_out_df['took_out'] == True]
        return took_out_df.sort_index(ascending=False)
        
    def calculate_tp_price_by_liquidity(self, symbol, position, df_liquidities, stop_loss_pct=2, tp_sl_ratio=1.5, offset=1) -> float:
        """_summary_
        计算止盈价格，根据流动性，做多则止盈价格在流动性之上，做空则止盈价格在流动性之下。
        Args:
            symbol (_type_): _description_
            position (_type_): _description_
            df_liquidities (_type_): _description_
            stop_loss_pct (int, optional): _description_. Defaults to 2.
            tp_sl_ratio (float, optional): _description_. Defaults to 1.5.
            offset (int, optional): _description_. Defaults to 1.

        Returns:
            float: _description_
        """        
        
        tp_price = 0.0
        # market_price = float(position['markPrice'])
        
        is_buy = position['side'] == 'long'
        price_col = 'Up' if is_buy else 'Dn'
        
        # sl_price = self.global_symbol_stop_loss_price.get(symbol, float(position['markPrice']))
        entry_price = float(position['entryPrice'])
        sl_price =  self.calculate_sl_price_by_pct(symbol, position, stop_loss_pct)
        threshold = 0.0
        if is_buy and sl_price > 0 :
            threshold = float(self.round_price_to_tick(symbol, (entry_price - sl_price ) * tp_sl_ratio + entry_price))
        elif not is_buy and sl_price > 0:
            threshold = float(self.round_price_to_tick(symbol, entry_price - (sl_price - entry_price ) * tp_sl_ratio))
         

        # 过滤有效的流动性价格
        valid_mask = df_liquidities[price_col] > threshold if is_buy else df_liquidities[price_col] < threshold
        df_valid_liquidities = df_liquidities[valid_mask]
        self.logger.debug(f"{symbol} : threshold={threshold} sl_price={sl_price} 有效的流动=\n {df_valid_liquidities[['timestamp','Up','Dn']]}")
        # 获取止盈价格
        tp_price = df_valid_liquidities.iloc[0][price_col] if len(df_valid_liquidities) > 0 else threshold
        if is_buy:
            tp_price = max(tp_price, threshold)
        else:
            tp_price = min(tp_price, threshold)
        tick_size = self.get_tick_size(symbol)
        
        # 计算止盈价格 , 做多则止盈价格在流动性之下tick_size，做空则止盈价格在流动性之上tick_size。
        if is_buy:
            tp_price = tp_price - offset * tick_size
        else:   
            tp_price = tp_price + offset * tick_size
            
        
        return tp_price
    
    @override
    def close_all_cache(self):
        super().close_all_cache()
        self.htf_liquidities.clear()
        self.global_symbol_take_profit_flag.clear()
        self.global_symbol_take_profit_price.clear()


    @override
    def reset_all_cache(self, symbol):
        super().reset_all_cache(symbol)
        self.htf_liquidities[symbol] = None
        self.global_symbol_take_profit_flag[symbol] = False
        self.global_symbol_take_profit_price[symbol] = None
           
    def set_tp_by_structure(self, symbol, position, pair_config, htf_Klines=None):
        """
        根据结构设置止盈
        """
        # 如果已经触发过全局止盈，则跳过 
        if self.global_symbol_take_profit_flag.get(symbol, False):            
            self.logger.info(f"{symbol} : 已经设置过全局止盈 tp_price={self.global_symbol_take_profit_price[symbol]}")
            return

        smc_strategy = pair_config.get('smc_strategy',{})
        htf = str(smc_strategy.get('HTF','15m')) 
        htf_prd = int(smc_strategy.get('HTF_swing_points_length',3))    
         
        # 寻找HTF的流动性，作为止盈位置    
        if htf_Klines is None:
            htf_Klines = self.get_historical_klines(symbol=symbol, bar=htf)
        htf_df = self.format_klines(htf_Klines)   
        htf_df_with_struct = self.build_struct(df=htf_df,prd=htf_prd)  

        # 寻找流动性
        htf_liquidity = self.htf_liquidities.get(symbol,None)
        if htf_liquidity is None:
            htf_liquidity = self.detect_liquidity_for_TP(symbol, htf_df_with_struct, position['side'], position['markPrice'])
            self.htf_liquidities[symbol] = htf_liquidity
        
        if len(htf_liquidity) <= 0:
            self.logger.info(f"{symbol} : 没有找到流动性，不设置止盈")
            return
       
        tp_price = self.calculate_tp_price_by_liquidity(symbol, position, htf_liquidity, self.stop_loss_pct, self.all_TP_SL_ratio)
        
        self.cancel_all_algo_orders(symbol=symbol, attachType='TP')
        
        if self.set_take_profit(symbol, position, tp_price):
            self.global_symbol_take_profit_flag[symbol] = True
            self.global_symbol_take_profit_price[symbol] = tp_price
            self.logger.info(f"{symbol} : [{position['side']}] 设置全局止盈价={tp_price}")
    
 
    def check_total_profit(self, symbol, position):
        """
        检查当前总盈利
        """
        pair_config = self.get_pair_config(symbol)  
        total_profit = self.calculate_average_profit(symbol, position)     
        cur_highest_total_profit = self.highest_total_profit.get(symbol, 0.0)    
        if total_profit > cur_highest_total_profit:            
            self.highest_total_profit[symbol] = total_profit
        profit_key = "盈利" if total_profit > 0.0 else "亏损"
        msg = f"{symbol} : {profit_key}={total_profit:.2f}% 方向={position['side']} 开仓={position['entryPrice']:.6f} 市价={position['markPrice']:.6f}"
        self.logger.info(msg)
        self.send_feishu_notification(msg)    
        
        # self.cancel_all_algo_orders(symbol=symbol) 
        
        smc_strategy = pair_config.get('smc_strategy',{})
        htf = str(smc_strategy.get('HTF','15m'))    
        htf_Klines = self.get_historical_klines(symbol=symbol, bar=htf)
          

        # 1. 根据总盈利设置止损
        self.set_sl_by_profit(symbol=symbol, position=position, profit=total_profit, pair_config=pair_config, kLines=htf_Klines)
        
        # 2. 根据结构设置止盈
        self.set_tp_by_structure(symbol=symbol, position=position, pair_config=pair_config, htf_Klines=htf_Klines)
        
        
        return
     



from typing import override
from taker.TrailingSLTaker import TrailingSLTaker
'''
自动设置移动止损单
'''
class TrailingSLAndTPTaker(TrailingSLTaker):
    def __init__(self,g_config, platform_config, feishu_webhook=None, monitor_interval=4,logger=None):
        super().__init__(g_config, platform_config, feishu_webhook, monitor_interval,logger)
        self.global_symbol_take_profit_flag = {} # 记录每个symbol是否设置全局止盈标志
        self.global_symbol_take_profit_price = {} # 记录每个symbol的止盈价格
        self.all_TP_SL_ratio = float(platform_config.get("all_TP_SL_ratio",1.5)) #The profit-loss ratio 盈亏比
        self.all_take_profit_pct = self.stop_loss_pct *  self.all_TP_SL_ratio
   
    def set_stop_loss_take_profit(self, symbol, position, stop_loss_price=None, take_profit_price=None) -> bool:
        if not stop_loss_price and not take_profit_price:
            self.logger.warning(f"{symbol}: No stop loss price or take profit price provided for {symbol}")
            return False   
        if not position:
            self.logger.warning(f"{symbol}: No position found for {symbol}")
            return False
        
        # 取消所有策略订单
        if_success = self.cancel_all_algo_orders(symbol=symbol)        
        
        if if_success:
            self.global_symbol_stop_loss_price[symbol] = None
            self.global_symbol_stop_loss_flag[symbol] = False
            return True
        
        if_stop_loss_success ,if_take_profit_success  = True , True     
        
        if stop_loss_price :
            if_stop_loss_success = self.set_stop_loss(symbol=symbol, position=position, stop_loss_price=stop_loss_price)
        if take_profit_price :
            if_take_profit_success = self.set_take_profit(symbol=symbol, position=position, take_profit_price=take_profit_price)
            
        is_successful =  if_stop_loss_success and if_take_profit_success
        
        order_take_profit_price = take_profit_price
        if take_profit_price is None:
            order_take_profit_price = self.calculate_take_profile_price(symbol, position, self.all_take_profit_pct)
            is_successful = self.set_take_profit(symbol, position, order_take_profit_price)    
        
        return is_successful
    
    @override
    def close_all_cache(self):
        super().close_all_cache()
        self.global_symbol_take_profit_flag.clear()
        self.global_symbol_take_profit_price.clear()
    @override    
    def reset_all_cache(self, symbol):
        self.reset_highest_profit_and_tier(symbol)
        self.reset_take_profie(symbol)
        self.global_symbol_take_profit_flag[symbol] = False
        self.global_symbol_take_profit_price[symbol] = None


import logging
import yaml
from logging.handlers import TimedRotatingFileHandler

from taker.TrailingSLTaker import TrailingSLTaker
from taker.TrailingSLAndTPTaker import TrailingSLAndTPTaker
from taker.ThreeLineTradingTaker import ThreeLineTradingTaker
from taker.SMCSLAndTPTaker import SMCSLAndTPTaker

def build_logger(log_config) -> logging.Logger:
            # 配置日志
        # log_file = "log/okx_MultiAssetNewTradingBot.log"
        log_file = log_config["file"] 
        logger = logging.getLogger(__name__)
        logger.setLevel(log_config["level"])

        file_handler = TimedRotatingFileHandler(log_file, when='midnight', interval=1, backupCount=7, encoding='utf-8')
        file_handler.suffix = "%Y-%m-%d"
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(lineno)d - %(levelname)s - %(message)s')
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

        console_handler = logging.StreamHandler()
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)
        
        return logger
    
def read_config_file(file_path):
    try:
        # 打开 YAML 文件
        with open(file_path, 'r', encoding='utf-8') as file:
            # 使用 yaml.safe_load 方法解析 YAML 文件内容
            data = yaml.safe_load(file)
            return data
    except FileNotFoundError:
        raise Exception(f"文件 {file_path} 未找到。")
    except yaml.YAMLError as e:
        raise Exception(f"解析 {file_path} 文件时出错: {e}")


def main():
    import importlib.metadata
    version = importlib.metadata.version("openfund-taker")
    
    # openfund_config_path = 'config.json'
    openfund_config_path = 'taker_config.yaml'
    config_data = read_config_file(openfund_config_path)
    
    
    # with open(openfund_config_path, 'r') as f:
    #     config_data = json.load(f)
        
    platform_config = config_data['okx']
    feishu_webhook_url = config_data['feishu_webhook']
    monitor_interval = config_data.get("monitor_interval", 60)  # 默认值为60秒
    logger = build_logger(config_data["Logger"])
    package_name = __package__ or "taker" 
    
    taker = config_data.get('actived_taker', 'SMCSLAndTPTaker')
        
    # 根据配置动态创建策略实例
    strategy_class = globals()[taker]
    bot = strategy_class(config_data, platform_config, feishu_webhook=feishu_webhook_url, monitor_interval=monitor_interval,logger=logger)
    logger.info(f" ++ {package_name}.{taker}:{version} is doing...")

    bot.monitor_total_profit()
    # bot = ThreeLineTradingBot(platform_config, feishu_webhook=feishu_webhook_url, monitor_interval=monitor_interval)
    # bot.monitor_klines()

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
import ccxt
import time
import logging
import requests
import json
from logging.handlers import TimedRotatingFileHandler


class CustomBitget(ccxt.bitget):
    def fetch(self, url, method='GET', headers=None, body=None):
        if headers is None:
            headers = {}
        headers['X-CHANNEL-API-CODE'] = 'tu3hz'
        return super().fetch(url, method, headers, body)


class MultiAssetTradingBot:
    def __init__(self, config, feishu_webhook=None, monitor_interval=4):
        self.leverage = float(config["leverage"])
        self.stop_loss_pct = config["stop_loss_pct"]
        self.low_trail_stop_loss_pct = config["low_trail_stop_loss_pct"]
        self.trail_stop_loss_pct = config["trail_stop_loss_pct"]
        self.higher_trail_stop_loss_pct = config["higher_trail_stop_loss_pct"]
        self.low_trail_profit_threshold = config["low_trail_profit_threshold"]
        self.first_trail_profit_threshold = config["first_trail_profit_threshold"]
        self.second_trail_profit_threshold = config["second_trail_profit_threshold"]
        self.feishu_webhook = feishu_webhook
        self.blacklist = set(config.get("blacklist", []))
        self.monitor_interval = monitor_interval  # 从配置文件读取的监控循环时间

        # 配置交易所
        self.exchange = CustomBitget({
            'apiKey': config["apiKey"],
            'secret': config["secret"],
            'password': config.get("password", ""),  # 如果 Bitget 需要 password，可以配置进去
            'timeout': 3000,
            'rateLimit': 50,
            'options': {'defaultType': 'swap'},
            # 'proxies': {'http': 'http://127.0.0.1:10100', 'https': 'http://127.0.0.1:10100'},
        })

        # 配置日志
        log_file = "log/multi_asset_bot.log"
        logger = logging.getLogger(__name__)
        logger.setLevel(logging.INFO)

        # 使用 TimedRotatingFileHandler 以天为单位进行日志分割
        file_handler = TimedRotatingFileHandler(log_file, when='midnight', interval=1, backupCount=7, encoding='utf-8')
        file_handler.suffix = "%Y-%m-%d"  # 设置日志文件名的后缀格式，例如 multi_asset_bot.log.2024-11-05
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

        console_handler = logging.StreamHandler()
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)

        self.logger = logger

        # 用于记录每个持仓的最高盈利值和当前档位
        self.highest_profits = {}
        self.current_tiers = {}
        self.detected_positions = {}
        # 检查持仓模式
        if not self.is_single_position_mode():
            self.logger.error("持仓模式无法双向持仓,可能是因为手上有持仓单子导致，请先平仓更改双向持仓后再运行程序。")
            self.send_feishu_notification("持仓模式无法双向持仓,可能是因为手上有持仓单子导致，请先平仓更改双向持仓后再运行程序。")
            raise SystemExit("持仓模式无法双向持仓,可能是因为手上有持仓单子导致，请先平仓更改双向持仓后再运行程序。")

    def is_single_position_mode(self):
        try:
            # 设置为双向持仓模式
            account_info = self.exchange.set_position_mode(hedged=True)

            # 获取 posMode 字段的值
            self.logger.info(f"程序启动，更改持仓模式为双向持仓")
            self.send_feishu_notification(f"程序启动，更改持仓模式为双向持仓")
            pos_mode = account_info.get('data', {}).get('posMode', None)

            # 如果 pos_mode 为 'single_mode'，则表示为单向持仓模式
            return pos_mode == 'hedge_mode'
        except Exception as e:
            self.logger.error(f"获取账户信息时出错: {e}")
            return False

    def send_feishu_notification(self, message):
        if self.feishu_webhook:
            try:
                headers = {'Content-Type': 'application/json'}
                payload = {"msg_type": "text", "content": {"text": message}}
                response = requests.post(self.feishu_webhook, json=payload, headers=headers)
                if response.status_code == 200:
                    self.logger.info("飞书通知发送成功")
                else:
                    self.logger.error("飞书通知发送失败，状态码: %s", response.status_code)
            except Exception as e:
                self.logger.error("发送飞书通知时出现异常: %s", str(e))

    def schedule_task(self):
        self.logger.info("启动主循环，开始执行任务调度...")
        try:
            while True:
                self.monitor_positions()
                time.sleep(self.monitor_interval)
        except KeyboardInterrupt:
            self.logger.info("程序收到中断信号，开始退出...")
        except Exception as e:
            error_message = f"程序异常退出: {str(e)}"
            self.logger.error(error_message)
            self.send_feishu_notification(error_message)

    def fetch_positions(self):
        try:
            positions = self.exchange.fetch_positions()
            return positions
        except Exception as e:
            self.logger.error(f"Error fetching positions: {e}")
            return []

    def close_position(self, symbol, side):
        try:
            # 获取当前持仓数量
            position = next((pos for pos in self.fetch_positions() if pos['symbol'] == symbol), None)
            if position is None or float(position['contracts']) == 0:
                self.logger.info(f"{symbol} 仓位已平，无需继续平仓")
                return True

            amount = float(position['contracts'])  # 使用当前持仓数量进行一次性清仓

            # 创建平仓订单，并加上 reduceOnly 参数

            bg_symbol = symbol.replace("/", "").replace(":USDT", "")

            # order = self.exchange.create_order(symbol, 'market', side, amount)
            order = self.exchange.privateMixPostV2MixOrderClosePositions({'symbol': bg_symbol, 'holdSide': side, 'productType': 'USDT-FUTURES'})

            if order['code'] == '00000' and order['data']['successList']:
                self.logger.info(f"Closed position for {symbol} with size {amount}, side: {side}")
                self.send_feishu_notification(f"Closed position for {symbol} with size {amount}, side: {side}")
                self.detected_positions.pop(symbol, None)
                self.highest_profits.pop(symbol, None)
                self.current_tiers.pop(symbol, None)
                return True
            else:
                self.logger.error(f"Failed to close position for {symbol}: {order}")
                return False
        except Exception as e:
            self.logger.error(f"Error closing position for {symbol}: {e}")
            return False

    def monitor_positions(self):
        positions = self.fetch_positions()
        current_symbols = set(position['symbol'] for position in positions if float(position['contracts']) != 0)

        closed_symbols = set(self.detected_positions.keys()) - current_symbols
        for symbol in closed_symbols:
            self.logger.info(f"手动平仓检测：{symbol} 已平仓，从监控中移除")
            self.send_feishu_notification(f"手动平仓检测：{symbol} 已平仓，从监控中移除")
            self.detected_positions.pop(symbol, None)

        for position in positions:
            symbol = position['symbol']
            position_amt = float(position['contracts'])
            entry_price = float(position['entryPrice'])
            current_price = float(position['markPrice'])
            side = position['side']
            td_mode = position['marginMode']

            if position_amt == 0:
                continue

            if symbol in self.blacklist:
                if symbol not in self.detected_positions:
                    self.send_feishu_notification(f"检测到黑名单品种：{symbol}，跳过监控")
                    self.detected_positions[symbol] = position_amt  # 临时存储持仓数量
                continue

            if symbol not in self.detected_positions:
                self.detected_positions[symbol] = position_amt
                self.highest_profits[symbol] = 0
                self.current_tiers[symbol] = "无"
                self.logger.info(
                    f"首次检测到仓位：{symbol}, 仓位数量: {position_amt}, 开仓价格: {entry_price}, 方向: {side}")
                self.send_feishu_notification(
                    f"首次检测到仓位：{symbol}, 仓位数量: {position_amt}, 开仓价格: {entry_price}, 方向: {side}")

            # 检测是否有新加仓
            if position_amt > self.detected_positions[symbol]:
                self.highest_profits[symbol] = 0  # 重置最高盈利
                self.current_tiers[symbol] = "无"  # 重置档位
                self.detected_positions[symbol] = position_amt
                self.logger.info(f"{symbol} 新仓检测到，重置最高盈利和档位。")
                continue  # 跳出当前循环

            if side == 'long':
                profit_pct = (current_price - entry_price) / entry_price * 100
            elif side == 'short':
                profit_pct = (entry_price - current_price) / entry_price * 100
            else:
                continue

            highest_profit = self.highest_profits.get(symbol, 0)
            if profit_pct > highest_profit:
                highest_profit = profit_pct
                self.highest_profits[symbol] = highest_profit

            current_tier = self.current_tiers.get(symbol, "无")
            if highest_profit >= self.second_trail_profit_threshold:
                current_tier = "第二档移动止盈"
            elif highest_profit >= self.first_trail_profit_threshold:
                current_tier = "第一档移动止盈"
            elif highest_profit >= self.low_trail_profit_threshold:
                current_tier = "低档保护止盈"
            else:
                current_tier = "无"

            self.current_tiers[symbol] = current_tier

            self.logger.info(
                f"监控 {symbol}，仓位: {position_amt}，方向: {side}，开仓价格: {entry_price}，当前价格: {current_price}，浮动盈亏: {profit_pct:.2f}%，最高盈亏: {highest_profit:.2f}%，当前档位: {current_tier}")

            if current_tier == "低档保护止盈":
                self.logger.info(f"回撤到{self.low_trail_stop_loss_pct:.2f}% 止盈")
                if profit_pct <= self.low_trail_stop_loss_pct:
                    self.logger.info(f"{symbol} 触发低档保护止盈，当前盈亏回撤到: {profit_pct:.2f}%，执行平仓")
                    self.close_position(symbol, side)
                    continue

            elif current_tier == "第一档移动止盈":
                trail_stop_loss = highest_profit * (1 - self.trail_stop_loss_pct)
                self.logger.info(f"回撤到 {trail_stop_loss:.2f}% 止盈")
                if profit_pct <= trail_stop_loss:
                    self.logger.info(
                        f"{symbol} 达到利润回撤阈值，当前档位：第一档移动止盈，最高盈亏: {highest_profit:.2f}%，当前盈亏: {profit_pct:.2f}%，执行平仓")
                    self.close_position(symbol, side)
                    continue

            elif current_tier == "第二档移动止盈":
                trail_stop_loss = highest_profit * (1 - self.higher_trail_stop_loss_pct)
                self.logger.info(f"回撤到 {trail_stop_loss:.2f}% 止盈")
                if profit_pct <= trail_stop_loss:
                    self.logger.info(
                        f"{symbol} 达到利润回撤阈值，当前档位：第二档移动止盈，最高盈亏: {highest_profit:.2f}%，当前盈亏: {profit_pct:.2f}%，执行平仓")
                    self.close_position(symbol, side)
                    continue

            if profit_pct <= -self.stop_loss_pct:
                self.logger.info(f"{symbol} 触发止损，当前盈亏: {profit_pct:.2f}%，执行平仓")
                self.close_position(symbol, side)


if __name__ == '__main__':
    with open('config.json', 'r') as f:
        config_data = json.load(f)

    # 选择交易平台，假设这里选择 Bitget
    platform_config = config_data['bitget']
    feishu_webhook_url = config_data['feishu_webhook']
    monitor_interval = config_data.get("monitor_interval", 4)  # 默认值为4秒

    bot = MultiAssetTradingBot(platform_config, feishu_webhook=feishu_webhook_url, monitor_interval=monitor_interval)
    bot.schedule_task()



# -*- coding: utf-8 -*-
import ccxt
import time
import requests
import traceback
import pandas as pd
import talib as ta
'''
自动设置移动止损单
'''
class TrailingSLTaker:
    def __init__(self,g_config, platform_config, feishu_webhook=None, monitor_interval=4,logger=None):
        self.trading_pairs_config = g_config.get('tradingPairs', {})
        
        
        self.stop_loss_pct = float(platform_config.get("all_stop_loss_pct",2))  # 全局止损百分比
        # 回撤比例
        self.low_trail_stop_loss_pct = platform_config["all_low_trail_stop_loss_pct"] # 第一档
        self.trail_stop_loss_pct = platform_config["all_trail_stop_loss_pct"]# 第二档
        self.higher_trail_stop_loss_pct = platform_config["all_higher_trail_stop_loss_pct"]# 第三档
        # 止损阈值
        self.low_trail_profit_threshold = platform_config["all_low_trail_profit_threshold"]# 第一档
        self.first_trail_profit_threshold = platform_config["all_first_trail_profit_threshold"]# 第二档
        self.second_trail_profit_threshold = platform_config["all_second_trail_profit_threshold"]# 第三档
        
        # 追踪止盈回撤
        self.open_trail_tp = platform_config.get("open_trail_tp",False)# 是否开启追踪止盈
        self.low_trail_tp_pct = platform_config["all_low_trail_tp_pct"]# 第一档
        self.trail_tp_pct = platform_config["all_trail_tp_pct"]# 第二档
        self.higher_trail_tp_pct = platform_config["all_higher_trail_tp_pct"]# 第三档
        
        self.feishu_webhook = feishu_webhook
        self.monitor_interval = monitor_interval  # 监控循环时间是分仓监控的3倍
        
        self.highest_total_profit = {}  # 记录最高总盈利
        
        # self.current_tier = {} # 记录当前的仓位模式
        
        self.positions_entry_price = {} # 记录每个symbol的开仓价格
                
        self.global_symbol_stop_loss_flag = {} # 记录每个symbol是否设置全局止损
        self.global_symbol_stop_loss_price = {} # 记录每个symbol的止损价格
        
        # 保留在止盈挂单中最高最低两个价格，计算止盈价格。
        self.max_market_price = {}
        self.min_market_price = {}
        
        self.cross_directions = {} # 持仓期间，存储每个交易对的交叉方向 

        # 配置交易所
        self.exchange = ccxt.okx({
            'apiKey': platform_config["apiKey"],
            'secret': platform_config["secret"],
            'password': platform_config["password"],
            'timeout': 3000,
            'rateLimit': 50,
            'options': {'defaultType': 'future'},
            'proxies': {'http': 'http://127.0.0.1:7890', 'https': 'http://127.0.0.1:7890'},
        })
        self.logger = logger
        self.position_mode = self.get_position_mode()  # 获取持仓模式
    
    def get_pair_config(self,symbol):
        return self.trading_pairs_config.get(symbol,{})
    # 获取市场信息
    def getMarket(self,symbol):
        self.exchange.load_markets()
        return self.exchange.market(symbol)
    # 获取tick_size
    def get_tick_size(self,symbol):
        return float(self.getMarket(symbol)['precision']['price'])
    # 获取价格精度
    def get_precision_length(self,symbol) -> int:
        tick_size = self.get_tick_size(symbol)
        return len(f"{tick_size:.10f}".rstrip('0').split('.')[1]) if '.' in f"{tick_size:.10f}" else 0
    # 获取当前持仓模式
    def get_position_mode(self):
        try:
            # 假设获取账户持仓模式的 API
            response = self.exchange.private_get_account_config()
            data = response.get('data', [])
            if data and isinstance(data, list):
                # 取列表的第一个元素（假设它是一个字典），然后获取 'posMode'
                position_mode = data[0].get('posMode', 'single')  # 默认值为单向
                self.logger.info(f"当前持仓模式: {position_mode}")
                return position_mode
            else:
                self.logger.error("无法检测持仓模式: 'data' 字段为空或格式不正确")
                return 'single'  # 返回默认值
        except Exception as e:
            self.logger.error(f"无法检测持仓模式: {e}")
            return None

    def send_feishu_notification(self, message):
        if self.feishu_webhook:
            try:
                headers = {'Content-Type': 'application/json'}
                payload = {"msg_type": "text", "content": {"text": message}}
                response = requests.post(self.feishu_webhook, json=payload, headers=headers)
                if response.status_code == 200:
                    self.logger.debug("飞书通知发送成功")
                else:
                    self.logger.warn("飞书通知发送失败，状态码: %s", response.status_code)
            except Exception as e:
                self.logger.error("发送飞书通知时出现异常: %s", str(e))

    def fetch_positions(self):
        try:
            positions = self.exchange.fetch_positions()
            return positions
        except Exception as e:
            self.logger.warning(f"Warn fetching positions: {e}")
            return []

    # 获取当前委托
    def fetch_open_orders(self,symbol,params={}):
        try:
            orders = self.exchange.fetch_open_orders(symbol=symbol,params=params)
            return orders
        except Exception as e:
            self.logger.error(f"Error fetching open orders: {e}")
            return None

    # 获取历史K线，不包含最新的一条
    def get_historical_klines_except_last(self,symbol, bar='1m', limit=241):
  
        klines = self.get_historical_klines(symbol, bar=bar, limit=limit)
        return klines[:-1]


    def get_historical_klines(self,symbol, bar='1m', limit=241):
        
        params = {
            # 'instId': instId,
        }
        klines = self.exchange.fetch_ohlcv(symbol, timeframe=bar,limit=limit,params=params)
        
        if klines :
            
            return klines
        else:
            raise ValueError("Unexpected response structure or missing candlestick data")

    def format_klines(self,klines) -> pd.DataFrame:
       
        """_summary_
            格式化K线数据
        Args:
            klines (_type_): _description_
        """
        klines_df = pd.DataFrame(klines, columns=['timestamp', 'open', 'high', 'low', 'close', 'volume']) 
        # 转换时间戳为日期时间
        klines_df['timestamp'] = pd.to_datetime(klines_df['timestamp'], unit='ms').dt.tz_localize('UTC').dt.tz_convert('Asia/Shanghai')
        
        return klines_df

    def is_pivot_high(self, data, index, period, check_bounds=False):
        """
        判断当前索引处是否为枢轴高点
        :param data: 包含 'high' 列的 DataFrame
        :param index: 当前索引
        :param period: 前后比较的周期数
        :param check_bounds: 是否检查边界
        :return: 是否为枢轴高点
        """
        if check_bounds and (index < period or index >= len(data) - period):
            return False
        current_high = data.at[index, 'high']
        prev_highs = data['high'].iloc[max(0,index - period):index]
        next_highs = data['high'].iloc[index :min(len(data),index + period + 1)]
        return all(current_high >= prev_highs) and all(current_high >= next_highs)


    def is_pivot_low(self, data, index, period, check_bounds=False):
        """
        判断当前索引处是否为枢轴低点
        :param data: 包含 'low' 列的 DataFrame
        :param index: 当前索引
        :param period: 前后比较的周期数
        :param check_bounds: 是否检查边界
        :return: 是否为枢轴低点
        """
        if check_bounds and (index < period or index >= len(data) - period):
            return False
        current_low = data.at[index, 'low']
        prev_lows = data['low'].iloc[max(0,index - period):index]
        next_lows = data['low'].iloc[index :min(len(data),index + period + 1)]
        return all(current_low <= prev_lows) and all(current_low <= next_lows)


    def get_last_solated_point(self,symbol,position,kLines ,prd=20):
        # 将K线数据转换为DataFrame格式    
        df = self.format_klines(kLines)
    
        # 根据position方向寻找孤立点
        side = position['side']
        window = prd  # 设置窗口大小，用于判断局部最值
        isolated_points = []
        for i in range(len(df)-1, -1, -1):

            if side == 'short':
                # 找最近的孤立高点
                if self.is_pivot_high(df, i, window):
                # 获取最近的孤立高点
                    isolated_points.append(df.iloc[i])
                    break
                
            else:
                # 找最近的孤立低点 
                if self.is_pivot_low(df, i, window):
                    isolated_points.append(df.iloc[i])
                    break
        
        return isolated_points


    def judge_correct_postion_side(self, symbol, pair_config, klines=None) -> str:

        '''
        零轴之上的macd与signal形成金叉
        零轴之下的死叉
        零轴之上的死叉-金叉-死叉
        零轴之下的金叉-死叉-金叉
        '''

        order_side = 'none'
        if 'macd_strategy' not in pair_config :
            return  order_side
        
        macd = pd.DataFrame(klines, columns=['timestamp', 'open', 'high', 'low', 'close', 'volume'])
        # 将时间戳转换为日期时间格式
        macd['timestamp'] = pd.to_datetime(macd['timestamp'], unit='ms').dt.strftime('%m-%d %H:%M')
            
        # 使用 TA-Lib 计算 MACD
        macd[['macd', 'signal', 'hist']] = pd.DataFrame(ta.MACD(macd['close'], fastperiod=12, slowperiod=26, signalperiod=9)).T
        # self.logger.debug(f"{symbol} : MACD Values = \n {macd.tail(5)}")
        
        
        # 计算最近三个交叉点
        last_up_crosses = []
        last_down_crosses = []
        other_crosses = []
        all_cross = []
        
        # 从最新K开始往前遍历
        for i in range(len(macd)-1, 2, -1):
            # 检查是否发生死叉（MACD从上方穿过Signal）
            if (macd.iloc[i-1]['macd'] <= macd.iloc[i-1]['signal'] and                 
                macd.iloc[i]['macd'] > macd.iloc[i]['signal'] 
                ):
                all_cross.append(('golden', i))
                
                # 判断如果都在零轴之上加入last_up_crosses , 判断如果都在零轴之下加入last_down_crosses
                if macd.iloc[i]['macd'] > 0 and macd.iloc[i]['signal'] > 0 :
                    last_up_crosses.append(('golden', i))
                elif macd.iloc[i]['macd'] < 0 and macd.iloc[i]['signal'] < 0 :
                    last_down_crosses.append(('golden', i))
                else:
                    other_crosses.append(('golden', i))
   
            # 检查是否发生死叉（MACD从上方穿过Signal）
            elif macd.iloc[i-1]['macd'] >= macd.iloc[i-1]['signal'] and macd.iloc[i]['macd'] < macd.iloc[i]['signal']:
                all_cross.append(('death', i))
                # 判断如果都在零轴之上加入last_up_crosses , 判断如果都在零轴之下加入last_down_crosses
                if macd.iloc[i]['macd'] > 0 and macd.iloc[i]['signal'] > 0 :
                    last_up_crosses.append(('death', i))
                elif macd.iloc[i]['macd'] < 0 and macd.iloc[i]['signal'] < 0 :
                    last_down_crosses.append(('death', i))
                else:
                    other_crosses.append(('golden', i))
            # 只保留最后三个交叉点
            if len(last_up_crosses) == 3 or len(last_down_crosses) == 3:
                break
            
        self.logger.debug(f"{symbol} : \n- 所有cross {all_cross} \n- 零轴之上cross {last_up_crosses} \n- 零轴之下cross {last_down_crosses} \n- 其他corss {other_crosses}。")
        
        # valid_klines = pair_config['macd_strategy'].get('valid_klines', 5)
        # 如果最新的交叉是金叉，且又是零轴上方的金叉
        if len(last_up_crosses) > 0 and all_cross[0][0] == 'golden' and all_cross[0][1] == last_up_crosses[0][1] :
            order_side =  'long'
            self.logger.debug(f"{symbol} : 零轴之上的macd与signal形成金叉{all_cross[0]} 。") 
            
        # 如果最新的交叉是死叉，且又是零轴下方的死叉
        elif len(last_down_crosses) > 0 and all_cross[0][0] == 'death' and all_cross[0][1] == last_down_crosses[0][1] :
            order_side ='short'
            self.logger.debug(f"{symbol} : 零轴之下的macd与signal形成死叉{all_cross[0]} 。")
        # 分析交叉点模式，要满足连续的三个交叉都是零上
        elif len(last_up_crosses) == 3 and len(all_cross) == 3:
      
           # 零轴之上的死叉-金叉-死叉模式
            if (last_up_crosses[0][0] == 'death' and 
                last_up_crosses[1][0] == 'golden' and 
                last_up_crosses[2][0] == 'death' 
                ):
                order_side = 'short'
                self.logger.debug(f"{symbol} : 零轴之上的死叉-金叉-死叉模式 {order_side}。")
            
        elif len(last_down_crosses) == 3 and len(all_cross) == 3:
            # 零轴之下的金叉-死叉-金叉模式
            if (last_down_crosses[0][0] == 'golden' and 
                  last_down_crosses[1][0] == 'death' and 
                  last_down_crosses[2][0] == 'golden' 
                  ):
                order_side = 'long'
                self.logger.debug(f"{symbol} : 零轴之下的金叉-死叉-金叉模式 {order_side}。")

        return order_side

    def is_profitable(self, symbol ,position) -> bool:
        # 判断是否进入浮赢阶段
        if position['side'] == 'long':
            # 多单判断持仓价格是否小于止损价格
            return position['entryPrice'] < self.global_symbol_stop_loss_price.get(symbol,0.0)
        else:
            # 空单判断持仓价格是否大于止损价格
            return position['entryPrice'] > self.global_symbol_stop_loss_price.get(symbol,0.0)

    def check_reverse_position(self,symbol,position,pair_config):
        if 'entryPrice' not in position or self.is_profitable(symbol,position):
            self.logger.debug(f"{symbol} : 方向={position['side']} 经进入浮赢阶段，不校验全局止损位置，止损价={self.global_symbol_stop_loss_price} ,开仓价={position['entryPrice']}。")
            return 
            
        side = position['side']
        try:

            klines_period = str(pair_config.get('klines_period', '1m'))
            klines = self.get_historical_klines_except_last(symbol=symbol,bar=klines_period)

            self.logger.debug(f"开始监控 {symbol} : klines {klines_period} - {len(klines)}")
            
            correct_side = self.judge_correct_postion_side(symbol=symbol, pair_config=pair_config, klines=klines)
            
            order_stop_loss_pct = None
            # 方向不一致 尽快平仓
            if correct_side != 'none' and correct_side != side :
                self.logger.info(f"{symbol}: 持仓方向={side} 与 正确方向={correct_side} 相反 , 减少止损。")
                order_stop_loss_pct = self.stop_loss_pct / 2
                self.logger.info(f"{symbol} 全局止损阈值-修正后= {self.stop_loss_pct:.2f} -> {order_stop_loss_pct:.2f}%")
            else :
                order_stop_loss_pct = self.stop_loss_pct
                self.logger.info(f"{symbol}: 持仓方向={side} 与 正确方向={correct_side} 相同 , 恢复正常。") 

            # 根据情况 重新修正止损         
            self.global_symbol_stop_loss_flag[symbol] = False  
           
            self.set_global_stop_loss(symbol=symbol,position=position, pair_config=pair_config, stop_loss_pct=order_stop_loss_pct)

                
        except Exception as e:
            error_message = f"程序异常退出: {str(e)}"
            self.logger.warning(error_message,exc_info=True)
            traceback.print_exc()
            self.send_feishu_notification(error_message)
    
    def calculate_sma_pandas(self,symbol,kLines,period):
        """
        使用 pandas 计算 SMA
        :param KLines K线
        :param period: SMA 周期
        :return: SMA 值
        """
 
        df = pd.DataFrame(kLines, columns=['timestamp', 'open', 'high', 'low', 'close', 'volume'])
        sma = df['close'].rolling(window=period).mean()
        return sma 
            
    def calculate_ema_pandas(self,symbol,kLines, period):
        """
        使用 pandas 计算 EMA
        :param KLines K线
        :param period: EMA 周期
        :return: EMA 值
        """

        df = pd.DataFrame(kLines, columns=['timestamp', 'open', 'high', 'low', 'close', 'volume'])
        # 计算EMA
        ema = df['close'].ewm(span=period, adjust=False).mean()
        return ema 
    
   # 计算平均利润
    def calculate_average_profit(self,symbol,position):
        # positions = self.fetch_positions()
        total_profit_pct = 0.0
        num_positions = 0

        entry_price = float(position['entryPrice'])
        current_price = float(position['markPrice'])
        side = position['side']

        # 计算单个仓位的浮动盈利百分比
        if side == 'long':
            profit_pct = (current_price - entry_price) / entry_price * 100
        elif side == 'short':
            profit_pct = (entry_price - current_price) / entry_price * 100
        else:
            return

        # 累加总盈利百分比
        total_profit_pct += profit_pct
        num_positions += 1

        # 记录单个仓位的盈利情况
        self.logger.info(f"仓位 {symbol}，方向: {side}，开仓价格: {entry_price}，当前价格: {current_price}，"
                            f"浮动盈亏: {profit_pct:.2f}%")

        # 计算平均浮动盈利百分比
        average_profit_pct = total_profit_pct / num_positions if num_positions > 0 else 0
        return average_profit_pct

    def reset_highest_profit_and_tier(self,symbol=None):
        """重置最高总盈利和当前档位状态"""
        if not symbol:
            self.highest_total_profit.clear()
        else :
            self.highest_total_profit[symbol] = 0.0

        # self.logger.debug("已重置最高总盈利")
  
    def reset_take_profie(self,symbol=None):
        if not symbol:
            self.global_symbol_stop_loss_price.clear()
            self.global_symbol_stop_loss_flag.clear()
            # 保留在止盈挂单中最高最低两个价格，计算止盈价格。
            self.max_market_price.clear()
            self.min_market_price.clear() 
            self.cross_directions.clear()
            
        else :
            # self.global_symbol_stop_loss_price[symbol] = 0.0
            self.global_symbol_stop_loss_flag[symbol] = False
            # 保留在止盈挂单中最高最低两个价格，计算止盈价格。
            self.max_market_price[symbol] = 0.0
            self.min_market_price[symbol] = float('inf')  # 初始化为浮点数最大值
            self.cross_directions[symbol] = None
       
    def round_price_to_tick(self,symbol, price):
        tick_size = float(self.exchange.market(symbol)['info']['tickSz'])
        # 计算 tick_size 的整数倍
        tick_decimals = len(f"{tick_size:.10f}".rstrip('0').split('.')[1]) if '.' in f"{tick_size:.10f}" else 0

        # 调整价格为 tick_size 的整数倍
        adjusted_price = round(price / tick_size) * tick_size
        return f"{adjusted_price:.{tick_decimals}f}"
        # 放弃当前委托
    
    def cancel_all_algo_orders(self, symbol, attachType=None) -> bool:
        """_summary_

        Args:
            symbol (_type_): _description_
            attachType (_type_, optional): "TP"|"SL". Defaults to None.
        """
        
        params = {
            "ordType": "conditional",
        }
        orders = self.fetch_open_orders(symbol=symbol,params=params)
        
        if orders is None:
            return False
        # 如果没有委托订单则直接返回
        if len(orders) == 0:           
            # self.global_symbol_stop_loss_flag[symbol] = False
            # self.logger.debug(f"{symbol} 未设置策略订单列表。")
            return True
     
        # algo_ids = [order['info']['algoId'] for order in orders if 'info' in order and 'algoId' in order['info']]
        algo_ids = []
        if attachType and attachType == 'SL':
            algo_ids = [order['id'] for order in orders if order['stopLossPrice'] and order['stopLossPrice'] > 0.0 ]
        elif attachType and attachType == 'TP':
            algo_ids = [order['id'] for order in orders if order['takeProfitPrice'] and order['takeProfitPrice'] > 0.0]
        else :
            algo_ids = [order['id'] for order in orders ]
            
        if len(algo_ids) == 0 :         
            self.logger.debug(f"{symbol} 未设置策略订单列表。")
            return True
        
        try:
            params = {
                "algoId": algo_ids,
                "trigger": 'trigger'
            }
            rs = self.exchange.cancel_orders(ids=algo_ids, symbol=symbol, params=params)
            
            return len(rs) > 0
            # self.global_symbol_stop_loss_flag[symbol] = False
            # self.logger.debug(f"Order {algo_ids} cancelled:{rs}")
        except Exception as e:
            self.logger.error(f"{symbol} Error cancelling order {algo_ids}: {e}")
            return False
      
    def set_stop_loss_take_profit(self, symbol, position, stop_loss_price=None,take_profit_price=None) -> bool:
        if not stop_loss_price :
            self.logger.warning(f"{symbol}: No stop loss price or take profit price provided for {symbol}")
            return False   
        if not position:
            self.logger.warning(f"{symbol}: No position found for {symbol}")
            return False
        
   
        if self.cancel_all_algo_orders(symbol=symbol, attachType='SL'):
            self.global_symbol_stop_loss_flag[symbol] = False
            # self.global_symbol_stop_loss_price[symbol] = 0.0
        else:
            return False

        return self.set_stop_loss(symbol=symbol, position=position, stop_loss_price=stop_loss_price) 
    

          
    def set_take_profit(self, symbol, position, take_profit_price=None) -> bool:

        # 计算下单数量
        amount = abs(float(position['contracts']))
    
        if amount <= 0:
            self.logger.warning(f"{symbol}: amount is 0 for {symbol}")
            return

        
       # 止损单逻辑 
        adjusted_price = float(self.round_price_to_tick(symbol, take_profit_price))
         
        tp_params = {
   
            
            'tpTriggerPx':adjusted_price,
            'tpOrdPx' : adjusted_price,
            'tpOrdKind': 'condition',
            'tpTriggerPxType':'last',
            
            'tdMode':position['marginMode'],
            'sz': str(amount),
            # 'closeFraction': '1',
            'cxlOnClosePos': True,
            'reduceOnly':True
        }
        
        side = 'short' 
        if position['side'] == side: # 和持仓反向相反下单
            side ='long'
            
        orderSide = 'buy' if side == 'long' else 'sell'
    
        max_retries = 3
        retry_count = 0

        while retry_count < max_retries:
            try:

                self.logger.debug(f"{symbol}: {orderSide} - TP at {adjusted_price} Starting....  ")
                
                self.exchange.create_order(
                    symbol=symbol,
                    type='optimal_limit_ioc',
  
                    price=adjusted_price,
                    side=orderSide,
                    amount=amount,
                    params=tp_params
                )
          
                self.logger.info(f"{symbol}: TP at {adjusted_price} Done.")
                break
                
               
            except ccxt.NetworkError as e:
                # 处理网络相关错误
                retry_count += 1
                self.logger.warning(f"!! 设置止盈单时发生网络错误,正在进行第{retry_count}次重试: {str(e)}")
                time.sleep(0.1)  # 重试前等待1秒
                continue
            except ccxt.ExchangeError as e:
                # 处理交易所API相关错误
                retry_count += 1
                self.logger.warning(f"!! 设置止盈单时发生交易所错误,正在进行第{retry_count}次重试: {str(e)}")
                time.sleep(0.1)
                continue
            except Exception as e:
                # 处理其他未预期的错误
                retry_count += 1
                self.logger.warning(f"!! 设置止盈单时发生未知错误,正在进行第{retry_count}次重试: {str(e)}")
                time.sleep(0.1)
                continue
            
        if retry_count >= max_retries:
            # 重试次数用完仍未成功设置止损单
            self.logger.warning(f"!! {symbol} 设置止盈单时重试次数用完仍未成功设置成功。 ")
            return False
        return True 
        

            
    def set_stop_loss(self, symbol, position, stop_loss_price=None , ord_type='conditional') -> bool:
        """
        设置止盈单
        :param symbol: 交易对
        :param position: 持仓信息
        :param stop_loss_price: 止损价格
        :param ord_type: 订单类型 ord_type: 'conditional'|'market'|'limit'
        :return: 是否成功设置止盈单
        """
        # 计算下单数量
        amount = abs(float(position['contracts']))
    
        if amount <= 0:
            self.logger.warning(f"{symbol}: amount is 0 for {symbol}")
            return

        # 止损单逻辑 
        adjusted_price = float(self.round_price_to_tick(symbol, stop_loss_price))
        
        # 默认市价止损，委托价格为-1时，执行市价止损。
        sl_params = {
            'slTriggerPx':adjusted_price , 
            'slOrdPx':'-1', # 委托价格为-1时，执行市价止损
            # 'slOrdPx' : adjusted_price,
            'slTriggerPxType':'last',
            'tdMode':position['marginMode'],
            'sz': str(amount),
            # 'closeFraction': '1',
            'cxlOnClosePos': True,
            'reduceOnly':True
        }
        if ord_type == 'limit':
            sl_params['slOrdPx'] = adjusted_price
        
        side = 'long' if position['side'] == 'short' else 'short' # 和持仓反向相反下单
        orderSide = 'buy' if side == 'long' else 'sell'
    
        max_retries = 3
        retry_count = 0

        while retry_count < max_retries:
            try:

                self.logger.debug(f"{symbol}: {orderSide} - Pre SL at {adjusted_price} Starting....  ")
                
                self.exchange.create_order(
                    symbol=symbol,
                    # type='optimal_limit_ioc',
                    type=ord_type,
                    # type='limit',
                    price=adjusted_price,
                    side=orderSide,
                    amount=amount,
                    params=sl_params
                )
                self.logger.info(f"{symbol}: SL at {adjusted_price} Done.")
                
                break               
               
            except ccxt.NetworkError as e:
                # 处理网络相关错误
                retry_count += 1
                self.logger.warning(f"!! 设置止损单时发生网络错误,正在进行第{retry_count}次重试: {str(e)}")
                time.sleep(0.1)  # 重试前等待1秒
                continue
            except ccxt.ExchangeError as e:
                # 处理交易所API相关错误
                retry_count += 1
                self.logger.warning(f"!! 设置止损单时发生交易所错误,正在进行第{retry_count}次重试: {str(e)}")
                time.sleep(0.1)
                continue
            except Exception as e:
                # 处理其他未预期的错误
                retry_count += 1
                self.logger.warning(f"!! 设置止损单时发生未知错误,正在进行第{retry_count}次重试: {str(e)}")
                time.sleep(0.1)
                continue
            
        if retry_count >= max_retries:
            # 重试次数用完仍未成功设置止损单
            self.logger.warning(f"!! {symbol} 设置止损单时重试次数用完仍未成功设置成功。 ")
            return False
        
        return True


    def calculate_sl_price_by_pct(self, symbol, position, sl_pct) -> float:
        side = position['side']
        if side == 'long':           
            sl_price = position['entryPrice'] * (1 - sl_pct/100)  
            
        elif side == 'short':        
            sl_price = position['entryPrice'] * (1 + sl_pct/100)
       
            
        sl_order_price = float(self.round_price_to_tick(symbol, sl_price)) 
        
        return sl_order_price   
            
    def set_global_stop_loss(self, symbol, position, pair_config , stop_loss_pct=None, kLines=None) -> bool:
        """设置全局止损
        
        Args:
            symbol: 交易对
            position: 持仓信息
            pair_config: 交易对配置
            
        """
        # 如果已经触发过全局止损并且有止损单，则跳过
        self.logger.debug(f"{symbol} : 是否设置过全局止损={self.global_symbol_stop_loss_flag.get(symbol, False)} global_sl_price={self.global_symbol_stop_loss_price.get(symbol, 0.0)}")  
        if self.global_symbol_stop_loss_flag.get(symbol, False):
            return
      
             
        if stop_loss_pct is None :
            stop_loss_pct = self.stop_loss_pct  
             
        # 根据持仓方向计算止损价格
        side = position['side']       
            
        sl_order_price = self.calculate_sl_price_by_pct(symbol, position, stop_loss_pct)
        # tp_order_price = float(self.round_price_to_tick(symbol, tp_price))
        
        # 验证止损价格是否合理20250312
        mark_price = float(position['markPrice'])
        if side == 'long':
            if sl_order_price >= mark_price:
                self.close_all_positions(symbol,position)
                self.logger.warning(f"{symbol}: !! 多头止损价格= {sl_order_price} 大于等于市场价格= {mark_price}，触发止损")
                return
        else:  # side == 'short'
            if sl_order_price <= mark_price:
                self.close_all_positions(symbol,position)
                self.logger.warning(f"{symbol}: !! 空头止损价格= {sl_order_price} 小于等于市场价格= {mark_price}，触发止损")
                return
                
        # 250228 没有指定止损回撤阈值stop_loss_pct...
        open_swing_point = pair_config.get('open_swing_point', False)
        if open_swing_point :
            if kLines is None :
                klines_period = str(pair_config.get('klines_period', '1m'))
                kLines = self.get_historical_klines(symbol=symbol,bar=klines_period)               
       
            swing_points_length = pair_config.get('swing_points_length', 20)
            isolated_points = self.get_last_solated_point(symbol ,position , kLines,prd=swing_points_length)  
        
            if len(isolated_points) > 0 :
                last_isolated_point = isolated_points[0]
                                    
                if side == 'long':
                
                    sl_order_price = min(last_isolated_point['low'],sl_order_price)
                else:
                
                    sl_order_price = max(last_isolated_point['high'],sl_order_price)
                    
                self.logger.debug(f"{symbol} : {side} 止损价={sl_order_price} 孤立点=\n{last_isolated_point} ")


        # has_set_sl = self.global_symbol_stop_loss_flag.get(symbol,False)
        last_sl_price= self.global_symbol_stop_loss_price.get(symbol,0.0)  
        if last_sl_price != 0.0 and last_sl_price == sl_order_price:
            self.global_symbol_stop_loss_flag[symbol] = True
            self.logger.debug(f"{symbol} : [{side}] 全局止损价没变化, {last_sl_price} = {sl_order_price}")
            return True
        
        if last_sl_price != 0.0:
            # 根据持仓方向判断是否需要设置止损
            if position['side'] == 'long':
                # 多仓时,新止损价需高于当前止损价
                if last_sl_price > sl_order_price :
                    self.logger.debug(f"{symbol} : 多仓止损价格 {sl_order_price} 低于当前止损价格 {last_sl_price}，不更新止损")
                    return True
                else:
                    self.logger.debug(f"{symbol} : 多仓止损价格 {sl_order_price} 高于当前止损价格 {last_sl_price}，更新止损")
            else:
                # 空仓时,新止损价需低于当前止损价
                if last_sl_price < sl_order_price :
                    self.logger.debug(f"{symbol} : 空仓止损价格 {sl_order_price} 高于当前止损价格 {last_sl_price}，不更新止损") 
                    return True
                else:
                    self.logger.debug(f"{symbol} : 空仓止损价格 {sl_order_price} 低于当前止损价格 {last_sl_price}，更新止损")
        else:
            self.logger.debug(f"{symbol} : 未设置过全局止损, 新止损价={sl_order_price}")
            
        try:
            # 设置止损单
            if_success = self.set_stop_loss_take_profit(
                symbol=symbol,
                position=position,
                stop_loss_price=sl_order_price
                # take_profit_price=tp_order_price
            )
            if if_success:
                # 设置全局止损标志
                self.logger.info(f"{symbol} : [{side}] 设置全局止损价={sl_order_price}")
                self.global_symbol_stop_loss_flag[symbol] = True
                self.global_symbol_stop_loss_price[symbol] = sl_order_price
                
            return if_success
                
        except Exception as e:
            error_msg = f"{symbol} : 设置止损时发生错误, {str(e)}"
            self.logger.error(error_msg)
            traceback.print_exc()
            self.send_feishu_notification(error_msg)  
    
    def calculate_take_profile_price(self, symbol, position, take_profile_pct, offset=1) -> float:
        tick_size = self.get_tick_size(symbol)

        entry_price = position['entryPrice']
        side = position['side']

        # 计算止盈价格。

        if side == 'long':
            take_profile_price = entry_price * (1+take_profile_pct/100) - offset * tick_size


        elif side == 'short':
        
            # base_price = entry_price * (1-take_profile_pct)
            take_profile_price = entry_price * (1-take_profile_pct/100) + offset * tick_size
     
        return float(self.round_price_to_tick(symbol,take_profile_price))
    
    # 计算回撤止盈价格
    def calculate_stop_loss_price(self, symbol, position, stop_loss_pct, offset=1) -> float:
        tick_size = self.get_tick_size(symbol)
        market_price = position['markPrice']
        entry_price = position['entryPrice']
        side = position['side']
        # base_price = abs(market_price-entry_price) * (1-stop_loss_pct)
        # 计算止盈价格，用市场价格（取持仓期间历史最高）减去开仓价格的利润，再乘以不同阶段的止盈百分比。
        latest_stop_loss_price = self.global_symbol_stop_loss_price.get(symbol,0.0) 
        if side == 'long':
            last_max_market_price = self.max_market_price.get(symbol,0.0)
            self.max_market_price[symbol] = max(market_price,last_max_market_price)
            base_price = abs(self.max_market_price[symbol] - entry_price) * (1-stop_loss_pct)
            stop_loss_price = entry_price + base_price - offset * tick_size
            if latest_stop_loss_price :
                stop_loss_price = max(stop_loss_price,latest_stop_loss_price)

        elif side == 'short':
            last_min_market_price = self.min_market_price.get(symbol,float('inf'))
            self.min_market_price[symbol] = min(market_price,last_min_market_price)
            base_price = abs(self.min_market_price[symbol] - entry_price) * (1-stop_loss_pct)
            stop_loss_price = entry_price - base_price + offset * tick_size
            if latest_stop_loss_price :
                stop_loss_price = min(stop_loss_price,latest_stop_loss_price)
        return float(self.round_price_to_tick(symbol,stop_loss_price))
    
    # 市价仓位平仓    
    def close_all_positions(self,symbol,position):

        amount = abs(float(position['contracts']))
        side = position['side']
        td_mode = position['marginMode']
        if amount > 0:
            try:
                self.logger.info(f"{symbol}: Preparing to close position for {symbol}, side: {side}, amount: {amount}")

                if self.position_mode == 'long_short_mode':
                    # 在双向持仓模式下，指定平仓方向
                    pos_side = 'long' if side == 'long' else 'short'
                else:
                    # 在单向模式下，不指定方向
                    pos_side = 'net'
                orderSide = 'buy' if side == 'long' else 'sell'
                
                
                params = {
                    
                    'mgnMode': td_mode,
                    'posSide': pos_side,
                    # 当市价全平时，平仓单是否需要自动撤销,默认为false. false：不自动撤单 true：自动撤单
                    'autoCxl': 'true'
            
                }

                # 发送平仓请求并获取返回值
                order = self.exchange.close_position(
                    symbol=symbol,
                    side=orderSide,
                    params=params
                )
                time.sleep(0.1)  # 短暂延迟后再试
                self.reset_take_profie(symbol)
                self.logger.info(f"{symbol} Close position response for {symbol}: {order}")
                self.send_feishu_notification(f"{symbol} 平仓订单完全成交 -{symbol} side: {side}")

            except Exception as e:
                self.logger.error(f"{symbol} Error closing position for {symbol}: {e}")
                self.send_feishu_notification(f"{symbol} Error closing position for {symbol}: {e}")             

    def check_stop_loss_trigger(self, symbol: str, position: dict) -> bool:
            """
            检查是否触发止损条件
            Args:
                symbol: 交易对
                position: 持仓信息
            Returns:
                bool: 是否需要平仓
            """
            latest_stop_loss_price = self.global_symbol_stop_loss_price.get(symbol,0.0)  
            if latest_stop_loss_price == 0.0:
                self.logger.warning(f"{symbol}: 未设置止损价格，执行平仓")
                return True
                
            mark_price = position['markPrice']
            side = position['side']
            
            sl_price = float(self.round_price_to_tick(symbol=symbol,price=latest_stop_loss_price))
            
            if side == 'long' and mark_price < sl_price:
                self.logger.warning(f"{'*'*60}\n[非正常关闭]: {symbol} 方向 {side} - 市场价格 {mark_price} 低于止盈 {latest_stop_loss_price}，触发全局止盈\n{'*'*60}")
                return True
            elif side == 'short' and mark_price > sl_price:
                self.logger.warning(f"{'*'*60}\n[非正常关闭]: {symbol} 方向 {side} - 市场价格 {mark_price} 高于止盈价 {latest_stop_loss_price}，触发全局止盈\n{'*'*60}")
                return True
                
            return False
    
    def check_position(self, symbol, position):
        # 清理趋势相反的仓位
        pair_config = self.trading_pairs_config.get(symbol, {})
        self.check_reverse_position(symbol=symbol, position=position, pair_config=pair_config)
        
        # 检查止损是否触发止盈
        # if self.check_stop_loss_trigger(symbol, position):
        #     self.close_all_positions(symbol=symbol, position=position)
        #     return
        
    
    def check_total_profit(self, symbol, position):
        """
        检查当前总盈利
        """
    
        total_profit = self.calculate_average_profit(symbol, position)
        
        msg = f"{symbol}: 盈利= {total_profit:.2f}% 开仓= {position['entryPrice']:.10f} 市价= {position['markPrice']:.10f}"
        self.logger.info(msg)
        self.send_feishu_notification(msg)
            
        cur_highest_total_profit = self.highest_total_profit.get(symbol, 0.0)    
        
        if total_profit > cur_highest_total_profit:
            cur_highest_total_profit = total_profit
            self.highest_total_profit[symbol] = total_profit
            
        current_tier = '无'   
        # 确定当前盈利档位
        if cur_highest_total_profit >= self.second_trail_profit_threshold:
            current_tier = "高档"
     
        elif cur_highest_total_profit>= self.first_trail_profit_threshold:
            current_tier = "中档"
         
        elif cur_highest_total_profit >= self.low_trail_profit_threshold:
            current_tier = "低档"
            
            
        if total_profit > 0.0 :
            self.logger.info(
                f"{symbol} 档位[{current_tier}]: 盈利: {total_profit:.2f}%，最高盈利: {cur_highest_total_profit:.2f}%")
                
        '''
        第一档:低档保护止盈:当盈利达到0.3%触发,要么到第二档,要么回到0.2%止盈
        第二档:盈利达到1%触发,记录最高价,最高价的80%是止盈位
        第三档:盈利达到3%触发,记录最高价,最高价的75%是止盈位
        
        '''
        # 各档止盈逻辑
        pair_config = self.get_pair_config(symbol)   
        # 根据不同档位设置止损价格,没有单独为交易对设置，用全局参数代替
        tier_config = {
            "低档": {
                "stop_loss_pct": float(pair_config.get('low_trail_stop_loss_pct',self.low_trail_stop_loss_pct)),
                "trail_tp_pct": float(pair_config.get('low_trail_tp_pct',self.low_trail_tp_pct))                
            },
            "中档": {
                "stop_loss_pct": float(pair_config.get('trail_stop_loss_pct',self.trail_stop_loss_pct)),
                "trail_tp_pct": float(pair_config.get('trail_tp_pct',self.trail_tp_pct))
            },
            "高档": {
                "stop_loss_pct": float(pair_config.get('higher_trail_stop_loss_pct',self.higher_trail_stop_loss_pct)),
                "trail_tp_pct": float(pair_config.get('higher_trail_tp_pct',self.higher_trail_tp_pct))
            }
        }

        if current_tier in tier_config:
            config = tier_config[current_tier]
            
            # 是否追踪止盈
            if self.open_trail_tp and total_profit < cur_highest_total_profit:
                cur_trail_tp_pct = round((1-total_profit/cur_highest_total_profit), 3)
                trail_tp_pct = float(config['trail_tp_pct'])
                if cur_trail_tp_pct >= trail_tp_pct:
                    self.close_all_positions(symbol=symbol, position=position)
                    msg = f"{symbol}: 盈利【{current_tier}】保护止盈={trail_tp_pct*100}%，价格回撤 -{cur_trail_tp_pct*100}%: 利润={total_profit:.2f}%，执行止盈平仓"
                    self.logger.info(msg)
                    self.send_feishu_notification(msg)
                    return

            # 记录日志
            self.logger.debug(f"{symbol}: 回撤止盈阈值: {config['stop_loss_pct']*100}%")
            
            # 计算回撤止损价格
            sl_price = self.calculate_stop_loss_price(
                symbol=symbol, 
                position=position,
                stop_loss_pct=config['stop_loss_pct']
            )
            
            # 检查价格是否变化
            latest_sl_price = self.global_symbol_stop_loss_price.get(symbol,0.0)
            # 检查止损价格是否有变化
            if sl_price == latest_sl_price:
                self.logger.debug(f"{symbol}: 回撤止损价格{latest_sl_price}未变化，不设置")
                return
            
            # 根据持仓方向判断是否需要更新止损价格
            should_update = False
            if position['side'] == 'long':
                # 多仓时,新止损价需高于当前止损价
                should_update = sl_price >= latest_sl_price
                if not should_update:
                    self.logger.debug(f"{symbol}: 多仓回撤止损价格{sl_price}低于当前止损价格{latest_sl_price}，不更新")
                else:
                    self.logger.debug(f"{symbol}: 多仓回撤止损价格{sl_price}高于当前止损价格{latest_sl_price}，更新")
            else:
                # 空仓时,新止损价需低于当前止损价
                should_update = sl_price <= latest_sl_price
                if not should_update:
                    self.logger.debug(f"{symbol}: 空仓回撤止损价格{sl_price}高于当前止损价格{latest_sl_price}，不更新")
                else:
                    self.logger.debug(f"{symbol}: 空仓回撤止损价格{sl_price}低于当前止损价格{latest_sl_price}，更新")
                    
            if not should_update:
                return
                
            # 设置止损
            if_success = self.set_stop_loss(symbol, position, stop_loss_price=sl_price)
            
            if if_success:
                # 更新回撤止损价格
         
                self.global_symbol_stop_loss_price[symbol] = sl_price
                self.global_symbol_stop_loss_flag[symbol] = True
                
                # 发送通知
                msg = (f"{symbol}: 盈利达到【{current_tier}】阈值，最高总盈利: {cur_highest_total_profit:.2f}%,"
                      f"当前盈利回撤到: {total_profit:.2f}%，市场价格:{position['markPrice']},"
                      f"设置回撤止损位: {sl_price:.9f}")
                self.logger.info(msg)
                self.send_feishu_notification(msg)
                
        else:

            # 默认全局止损
           
            self.set_global_stop_loss(symbol, position, pair_config)
            self.logger.info(f"{symbol}: 全局止损阈值: {self.stop_loss_pct:.2f}%")
            
        return
    
    def close_all_cache(self):
        self.reset_highest_profit_and_tier()
        self.reset_take_profie()
        self.positions_entry_price= {}
    
    def reset_all_cache(self, symbol):
        self.reset_highest_profit_and_tier(symbol)
        self.reset_take_profie(symbol)
        self.global_symbol_stop_loss_price[symbol] = 0.0
        self.global_symbol_stop_loss_flag[symbol] = False
        
    def monitor_total_profit(self):
        self.logger.info("启动主循环，开始监控总盈利...")
        # previous_position_size = sum(
        #     abs(float(position['contracts'])) for position in self.fetch_positions())  # 初始总仓位大小
        while True:
            
            # 检查是否有仓位
            try:
                positions = self.fetch_positions()
        
            except Exception as e:
                self.logger.warning(f"获取仓位信息时发生错误: {e}")
                time.sleep(1)
                continue
            
            try:
                
                self.logger.info("+" * 60)

                for position in positions:
                    symbol = position['symbol']
                    cur_entry_price = position['entryPrice']
                    # 检查仓位是否有变化 TODO 永远不会执行。
                    if symbol in self.positions_entry_price and cur_entry_price != self.positions_entry_price[symbol]:
                        # 新开仓
                        self.reset_all_cache(symbol)
                    
                    if symbol not in self.positions_entry_price:
                        self.positions_entry_price[symbol] = cur_entry_price
                        msg = f"{symbol} : ## 重新开仓。 入场方向={position['side']} 入场价格={cur_entry_price} ##"
                        self.logger.info(msg)
                        self.send_feishu_notification(msg)
                        
                    self.check_total_profit(symbol, position)
                    time.sleep(0.1) 
                    # 检查仓位和挂单是否有问题
                    self.check_position(symbol, position)

                self.logger.info("-" * 60)
                time.sleep(self.monitor_interval)

            except Exception as e:
                # print(e)
                error_message = f"程序异常退出: {str(e)}"
                traceback.print_exc()
                self.logger.error(error_message)
                self.send_feishu_notification(error_message)
                continue
            except KeyboardInterrupt:
                self.logger.info("程序收到中断信号，开始退出...")
                break

